@@ -16,7 +16,7 @@
 #ifndef RADIUSD_H
 #define RADIUSD_H
 /**
- * $Id: b2a0a0f6424f73d9f5c0bc5032ae4ee3a64b65bc $
+ * $Id: 094af09c1db2c7c068e6e602798f985c905b07cb $
  *
  * @file radiusd.h
  * @brief Structures, prototypes and global variables for the FreeRADIUS server.
@@ -24,7 +24,7 @@
  * @copyright 1999-2000,2002-2008  The FreeRADIUS server project
  */
 
-RCSIDH(radiusd_h, "$Id: b2a0a0f6424f73d9f5c0bc5032ae4ee3a64b65bc $")
+RCSIDH(radiusd_h, "$Id: 094af09c1db2c7c068e6e602798f985c905b07cb $")
 
 #include <freeradius-devel/libradius.h>
 #include <freeradius-devel/radpaths.h>
@@ -114,6 +114,7 @@ typedef struct main_config {
 	fr_ipaddr_t	myip;				//!< IP to bind to. Set on command line.
 	uint16_t	port;				//!< Port to bind to. Set on command line.
 
+	bool		suppress_secrets;		//!< for debug levels < 3
 	bool		log_auth;			//!< Log all authentication attempts.
 	bool		log_accept;			//!< Log Access-Accept
 	bool		log_reject;			//!< Log Access-Reject
@@ -140,6 +141,8 @@ typedef struct main_config {
 	uint32_t	cleanup_delay;			//!< How long before cleaning up cached responses.
 	uint32_t	max_requests;
 
+	bool		postauth_client_lost;		//!< Whether to run Post-Auth-Type Client-Lost section
+
 	uint32_t	debug_level;
 	char const	*log_file;
 	int		syslog_facility;
@@ -312,8 +315,9 @@ struct rad_request {
 #define RAD_REQUEST_LVL_DEBUG3	(3)
 #define RAD_REQUEST_LVL_DEBUG4	(4)
 
-#define RAD_REQUEST_OPTION_COA	(1 << 0)
-#define RAD_REQUEST_OPTION_CTX	(1 << 1)
+#define RAD_REQUEST_OPTION_COA		(1 << 0)
+#define RAD_REQUEST_OPTION_CTX 		(1 << 1)
+#define RAD_REQUEST_OPTION_CANCELLED	(1 << 2)
 
 #define SECONDS_PER_DAY		86400
 #define MAX_REQUEST_TIME	30
@@ -374,8 +378,8 @@ int		rad_accounting(REQUEST *);
 int		rad_coa_recv(REQUEST *request);
 
 /* session.c */
-int		rad_check_ts(uint32_t nasaddr, uint32_t nas_port, char const *user, char const *sessionid);
-int		session_zap(REQUEST *request, uint32_t nasaddr,
+int		rad_check_ts(fr_ipaddr_t const *nas_addr, uint32_t nas_port, char const *user, char const *sessionid);
+int		session_zap(REQUEST *request, fr_ipaddr_t const *nas_addr,
 			    uint32_t nas_port, char const *user,
 			    char const *sessionid, uint32_t cliaddr,
 			    char proto, int session_time);
@@ -583,7 +587,7 @@ void radius_event_free(void);
 int radius_event_process(void);
 void radius_update_listener(rad_listen_t *listener);
 void revive_home_server(void *ctx);
-void mark_home_server_dead(home_server_t *home, struct timeval *when);
+void mark_home_server_dead(home_server_t *home, struct timeval *when, bool down);
 
 /* evaluate.c */
 typedef struct fr_cond_t fr_cond_t;
@@ -603,6 +607,13 @@ int dual_tls_recv(rad_listen_t *listener);
 int dual_tls_send(rad_listen_t *listener, REQUEST *request);
 int proxy_tls_recv(rad_listen_t *listener);
 int proxy_tls_send(rad_listen_t *listener, REQUEST *request);
+#ifdef WITH_COA_TUNNEL
+int proxy_tls_send_reply(rad_listen_t *listener, REQUEST *request);
+int dual_tls_send_coa_request(rad_listen_t *listener, REQUEST *request);
+void listen_coa_free(void);
+void listen_coa_add(rad_listen_t *listener, char const *key);
+int listen_coa_find(REQUEST *request, char const *key);
+#endif
 #endif
 
 /*
