@@ -1,7 +1,7 @@
 /*
  * realms.c	Realm handling code
  *
- * Version:     $Id: de0f18586de777e679cd9c5958a54f8fa91173c7 $
+ * Version:     $Id: 0a3b8f94b86b0590951fa693198c9b702d5d8d29 $
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
  * Copyright 2007  Alan DeKok <aland@deployingradius.com>
  */
 
-RCSID("$Id: de0f18586de777e679cd9c5958a54f8fa91173c7 $")
+RCSID("$Id: 0a3b8f94b86b0590951fa693198c9b702d5d8d29 $")
 
 #include <freeradius-devel/radiusd.h>
 #include <freeradius-devel/realms.h>
@@ -37,6 +37,10 @@ static rbtree_t *realms_byname = NULL;
 bool home_servers_udp = false;
 #endif
 
+#ifdef HAVE_DIRENT_H
+#include <dirent.h>
+#endif
+
 #ifdef HAVE_REGEX
 typedef struct realm_regex realm_regex_t;
 
@@ -53,6 +57,9 @@ static realm_regex_t *realms_regex = NULL;
 
 struct realm_config {
 	CONF_SECTION		*cs;
+#ifdef HAVE_DIRENT_H
+	char const		*directory;
+#endif
 	uint32_t		dead_time;
 	uint32_t		retry_count;
 	uint32_t		retry_delay;
@@ -66,6 +73,10 @@ static const FR_NAME_NUMBER home_server_types[] = {
 	{ "acct",		HOME_TYPE_ACCT },
 	{ "auth+acct",		HOME_TYPE_AUTH_ACCT },
 	{ "coa",		HOME_TYPE_COA },
+#ifdef WITH_COA_TUNNEL
+	{ "auth+coa",		HOME_TYPE_AUTH_COA },
+	{ "auth+acct+coa",	HOME_TYPE_AUTH_ACCT_COA },
+#endif
 	{ NULL, 0 }
 };
 
@@ -89,7 +100,7 @@ static realm_config_t *realm_config = NULL;
 static rbtree_t	*home_servers_byaddr = NULL;
 static rbtree_t	*home_servers_byname = NULL;
 #ifdef WITH_STATS
-static int home_server_max_number = 0;
+int home_server_max_number = 0;
 static rbtree_t	*home_servers_bynumber = NULL;
 #endif
 
@@ -107,6 +118,10 @@ static const CONF_PARSER proxy_config[] = {
 
 	{ "dynamic", FR_CONF_OFFSET(PW_TYPE_BOOLEAN, realm_config_t, dynamic), NULL },
 
+#ifdef HAVE_DIRENT_H
+	{ "directory", FR_CONF_OFFSET(PW_TYPE_STRING, realm_config_t, directory), NULL },
+#endif
+
 	{ "dead_time", FR_CONF_OFFSET(PW_TYPE_INTEGER, realm_config_t, dead_time), STRINGIFY(DEAD_TIME)  },
 
 	{ "wake_all_if_all_dead", FR_CONF_OFFSET(PW_TYPE_BOOLEAN, realm_config_t, wake_all_if_all_dead), "no" },
@@ -148,12 +163,12 @@ static int home_server_addr_cmp(void const *one, void const *two)
 	home_server_t const *a = one;
 	home_server_t const *b = two;
 
-	if (a->server && !b->server) return -1;
-	if (!a->server && b->server) return +1;
-	if (a->server && b->server) {
+	if (a->virtual_server && !b->virtual_server) return -1;
+	if (!a->virtual_server && b->virtual_server) return +1;
+	if (a->virtual_server && b->virtual_server) {
 		rcode = a->type - b->type;
 		if (rcode != 0) return rcode;
-		return strcmp(a->server, b->server);
+		return strcmp(a->virtual_server, b->virtual_server);
 	}
 
 	if (a->port < b->port) return -1;
@@ -264,6 +279,14 @@ static ssize_t xlat_home_server(UNUSED void *instance, REQUEST *request,
 			state = "dead";
 			break;
 
+		case HOME_STATE_CONNECTION_FAIL:
+			state = "fail";
+			break;
+
+		case HOME_STATE_ADMIN_DOWN:
+			state = "down";
+			break;
+
 		default:
 			state = "unknown";
 			break;
@@ -313,6 +336,60 @@ static ssize_t xlat_server_pool(UNUSED void *instance, REQUEST *request,
 
 	return xlat_cs(request->home_pool->cs, fmt, out, outlen);
 }
+
+
+/*
+ *	Xlat for %{home_server_dynamic:foo}
+ */
+static ssize_t xlat_home_server_dynamic(UNUSED void *instance, REQUEST *request,
+					char const *fmt, char *out, size_t outlen)
+{
+	int type;
+	char const *p;
+	home_server_t *home;
+
+	if (outlen < 2) return 0;
+
+	switch (request->packet->code) {
+	case PW_CODE_ACCESS_REQUEST:
+		type = HOME_TYPE_AUTH;
+		break;
+
+#ifdef WITH_ACCOUNTING
+	case PW_CODE_ACCOUNTING_REQUEST:
+		type = HOME_TYPE_ACCT;
+		break;
+#endif
+
+#ifdef WITH_COA
+	case PW_CODE_COA_REQUEST:
+	case PW_CODE_DISCONNECT_REQUEST:
+		type = HOME_TYPE_COA;
+		break;
+#endif
+
+	default:
+		*out = '\0';
+		return 0;
+	}
+
+	p = fmt;
+	while (isspace((int) *p)) p++;
+
+	home = home_server_byname(p, type);
+	if (!home) {
+		*out = '\0';
+		return 0;
+	}
+
+	/*
+	 *	1 for dynamic, 0 for static
+	 */
+	out[0] = '0' + home->dynamic;
+	out[1] = '\0';
+
+	return 1;
+}
 #endif
 
 void realms_free(void)
@@ -360,13 +437,23 @@ static CONF_PARSER home_server_coa[] = {
 	{ "mrd",  FR_CONF_OFFSET(PW_TYPE_INTEGER, home_server_t, coa_mrd), STRINGIFY(30) },
 	CONF_PARSER_TERMINATOR
 };
+
+
+
+#ifdef WITH_COA_TUNNEL
+static CONF_PARSER home_server_recv_coa[] = {
+	{ "virtual_server",  FR_CONF_OFFSET(PW_TYPE_STRING, home_server_t, recv_coa_server), NULL },
+	CONF_PARSER_TERMINATOR
+};
+#endif
+
 #endif
 
 static CONF_PARSER home_server_config[] = {
 	{ "ipaddr", FR_CONF_OFFSET(PW_TYPE_COMBO_IP_ADDR, home_server_t, ipaddr), NULL },
 	{ "ipv4addr", FR_CONF_OFFSET(PW_TYPE_IPV4_ADDR, home_server_t, ipaddr), NULL },
 	{ "ipv6addr", FR_CONF_OFFSET(PW_TYPE_IPV6_ADDR, home_server_t, ipaddr), NULL },
-	{ "virtual_server", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_NOT_EMPTY, home_server_t, server), NULL },
+	{ "virtual_server", FR_CONF_OFFSET(PW_TYPE_STRING | PW_TYPE_NOT_EMPTY, home_server_t, virtual_server), NULL },
 
 	{ "port", FR_CONF_OFFSET(PW_TYPE_SHORT, home_server_t, port), "0" },
 
@@ -409,6 +496,9 @@ static CONF_PARSER home_server_config[] = {
 
 #ifdef WITH_COA
 	{ "coa", FR_CONF_POINTER(PW_TYPE_SUBSECTION, NULL), (void const *) home_server_coa },
+#ifdef WITH_COA_TUNNEL
+	{ "recv_coa", FR_CONF_POINTER(PW_TYPE_SUBSECTION, NULL), (void const *) home_server_recv_coa },
+#endif
 #endif
 
 	CONF_PARSER_TERMINATOR
@@ -458,7 +548,7 @@ void realm_home_server_sanitize(home_server_t *home, CONF_SECTION *cs)
 	FR_INTEGER_BOUND_CHECK("check_timeout", home->ping_timeout, >=, 1);
 	FR_INTEGER_BOUND_CHECK("check_timeout", home->ping_timeout, <=, 10);
 
-	FR_INTEGER_BOUND_CHECK("revive_interval", home->revive_interval, >=, 60);
+	FR_INTEGER_BOUND_CHECK("revive_interval", home->revive_interval, >=, 10);
 	FR_INTEGER_BOUND_CHECK("revive_interval", home->revive_interval, <=, 3600);
 
 #ifdef WITH_COA
@@ -515,7 +605,7 @@ static bool home_server_insert(home_server_t *home, CONF_SECTION *cs)
 		return false;
 	}
 
-	if (!home->server && !rbtree_insert(home_servers_byaddr, home)) {
+	if (!home->virtual_server && !rbtree_insert(home_servers_byaddr, home)) {
 		rbtree_deletebydata(home_servers_byname, home);
 		cf_log_err_cs(cs, "Internal error %d adding home server %s", __LINE__, home->log_name);
 		return false;
@@ -557,7 +647,7 @@ bool realm_home_server_add(home_server_t *home)
 		return false;
 	}
 
-	if (!home->server && (rbtree_finddata(home_servers_byaddr, home) != NULL)) {
+	if (!home->virtual_server && (rbtree_finddata(home_servers_byaddr, home) != NULL)) {
 		char buffer[INET6_ADDRSTRLEN + 3];
 
 		inet_ntop(home->ipaddr.af, &home->ipaddr.ipaddr, buffer, sizeof(buffer));
@@ -608,6 +698,20 @@ bool realm_home_server_add(home_server_t *home)
 		}
 	}
 
+#ifdef WITH_COA_TUNNEL
+	if (home->recv_coa) {
+		if (!home->tls) {
+			ERROR("TLS is required in order to accept CoA requests from a home server");
+			return false;
+		}
+
+		if (!home->recv_coa_server) {
+			ERROR("A 'virtual_server' configuration is required in order to accept CoA requests from a home server");
+			return false;
+		}
+	}
+#endif
+
 	/*
 	 *	Mark it as already processed
 	 */
@@ -666,23 +770,25 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 	} else if (cf_pair_find(cs, "virtual_server") != NULL) {
 		home->ipaddr.af = AF_UNSPEC;	/* mark ipaddr as unused */
 
-		if (!home->server) {
+		if (!home->virtual_server) {
 			cf_log_err_cs(cs, "Invalid value for virtual_server");
 			goto error;
 		}
 
 		/*
-		 *	Try and find a 'server' section off the root of
+		 *	Try and find a "server" section off the root of
 		 *	the config with a name that matches the
 		 *	virtual_server.
 		 */
-		if (!cf_section_sub_find_name2(rc->cs, "server", home->server)) {
-			cf_log_err_cs(cs, "No such server %s", home->server);
+		if (!rc) goto error;
+
+		if (!cf_section_sub_find_name2(rc->cs, "server", home->virtual_server)) {
+			cf_log_err_cs(cs, "No such server %s", home->virtual_server);
 			goto error;
 		}
 
 		home->secret = "";
-		home->log_name = talloc_typed_strdup(home, home->server);
+		home->log_name = talloc_typed_strdup(home, home->virtual_server);
 	/*
 	 *	Otherwise it's an invalid config section and we
 	 *	raise an error.
@@ -713,11 +819,22 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 
 #ifdef WITH_COA
  		case HOME_TYPE_COA:
-			if (home->server != NULL) {
+			if (home->virtual_server != NULL) {
 				cf_log_err_cs(cs, "Home servers of type \"coa\" cannot point to a virtual server");
 				goto error;
 			}
 			break;
+
+#ifdef WITH_COA_TUNNEL
+		case HOME_TYPE_AUTH_ACCT_COA:
+			home->dual = true;
+			home->recv_coa = true;
+			break;
+
+		case HOME_TYPE_AUTH_COA:
+			home->recv_coa = true;
+			break;
+#endif
 #endif
 
   		case HOME_TYPE_INVALID:
@@ -744,6 +861,10 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 			}
 
 			if (((home->type == HOME_TYPE_AUTH) ||
+#ifdef WITH_COA_TUNNEL
+			     (home->type == HOME_TYPE_AUTH_COA) ||
+			     (home->type == HOME_TYPE_AUTH_ACCT_COA) ||
+#endif
 			     (home->type == HOME_TYPE_AUTH_ACCT)) && !home->ping_user_password) {
 				cf_log_err_cs(cs, "You must supply a 'password' to enable status_check=request");
 				goto error;
@@ -777,8 +898,9 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 			cf_log_err_cs(cs, "Server not built with support for RADIUS over TCP");
 			goto error;
 #endif
-			if (home->ping_check != HOME_PING_CHECK_NONE) {
-				cf_log_err_cs(cs, "Only 'status_check = none' is allowed for home "
+			if ((home->ping_check != HOME_PING_CHECK_NONE) &&
+			    (home->ping_check != HOME_PING_CHECK_STATUS_SERVER)) {
+				cf_log_err_cs(cs, "Only 'status_check = status-server' is allowed for home "
 					      "servers with 'proto = tcp'");
 				goto error;
 			}
@@ -792,7 +914,7 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 		home->proto = proto;
 	}
 
-	if (!home->server && rbtree_finddata(home_servers_byaddr, home)) {
+	if (!home->virtual_server && rbtree_finddata(home_servers_byaddr, home)) {
 		cf_log_err_cs(cs, "Duplicate home server");
 		goto error;
 	}
@@ -808,6 +930,38 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 	}
 #endif
 
+	/*
+	 *	Check the reverse CoA configuration.
+	 */
+#ifdef WITH_COA_TUNNEL
+	if (home->recv_coa) {
+		if (!tls) {
+			ERROR("TLS is required in order to accept CoA requests from a home server");
+			goto error;
+		}
+
+		if (!home->recv_coa_server) {
+			ERROR("A 'virtual_server' configuration is required in order to accept CoA requests from a home server");
+			goto error;
+		}
+
+		/*
+		 *	Try and find a 'server' section off the root of
+		 *	the config with a name that matches the coa
+		 *	virtual_server.
+		 */
+		if (!rc) {
+			ERROR("Dynamic home servers cannot accept CoA requests");
+			goto error;
+		}
+
+		if (!cf_section_sub_find_name2(rc->cs, "server", home->recv_coa_server)) {
+			cf_log_err_cs(cs, "No such coa server %s", home->recv_coa_server);
+			goto error;
+		}
+	}
+#endif
+
 	/*
 	 *	If were doing RADSEC (tls+tcp) the secret should default
 	 *	to radsec, else a secret must be set.
@@ -827,7 +981,7 @@ home_server_t *home_server_afrom_cs(TALLOC_CTX *ctx, realm_config_t *rc, CONF_SE
 	/*
 	 *	Virtual servers have some TLS restrictions.
 	 */
-	if (home->server) {
+	if (home->virtual_server) {
 		if (tls) {
 			cf_log_err_cs(cs, "Virtual home_servers cannot have a \"tls\" subsection");
 			goto error;
@@ -1052,6 +1206,16 @@ static int pool_check_home_server(UNUSED realm_config_t *rc, CONF_PAIR *cp,
 	case HOME_TYPE_ACCT:
 		myhome.type = HOME_TYPE_AUTH_ACCT;
 		home = rbtree_finddata(home_servers_byname, &myhome);
+#ifdef WITH_COA_TUNNEL
+		if (!home) {
+			myhome.type = HOME_TYPE_AUTH_COA;
+			home = rbtree_finddata(home_servers_byname, &myhome);
+			if(!home) {
+				myhome.type = HOME_TYPE_AUTH_ACCT_COA;
+				home = rbtree_finddata(home_servers_byname, &myhome);
+			}
+		}
+#endif
 		if (home) {
 			*phome = home;
 			return 1;
@@ -1261,7 +1425,7 @@ static int server_pool_add(realm_config_t *rc,
 			goto error;
 		}
 
-		if (!pool->fallback->server) {
+		if (!pool->fallback->virtual_server) {
 			cf_log_err_cs(cs, "Fallback home_server %s does NOT contain a virtual_server directive",
 				      pool->fallback->log_name);
 			goto error;
@@ -1343,6 +1507,16 @@ static int server_pool_add(realm_config_t *rc,
 			case HOME_TYPE_ACCT:
 				myhome.type = HOME_TYPE_AUTH_ACCT;
 				home = rbtree_finddata(home_servers_byname, &myhome);
+#ifdef WITH_COA_TUNNEL
+				if (!home) {
+					myhome.type = HOME_TYPE_AUTH_COA;
+					home = rbtree_finddata(home_servers_byname, &myhome);
+					if (!home) {
+						myhome.type = HOME_TYPE_AUTH_ACCT_COA;
+						home = rbtree_finddata(home_servers_byname, &myhome);
+					}
+				}
+#endif
 				break;
 
 			default:
@@ -1537,7 +1711,7 @@ static int old_server_add(realm_config_t *rc, CONF_SECTION *cs,
 			home->src_ipaddr.af = home->ipaddr.af;
 		} else {
 			home->ipaddr.af = AF_UNSPEC;
-			home->server = server;
+			home->virtual_server = server;
 		}
 		talloc_free(q);
 
@@ -2251,9 +2425,68 @@ int realms_init(CONF_SECTION *config)
 #ifdef WITH_PROXY
 	xlat_register("home_server", xlat_home_server, NULL, NULL);
 	xlat_register("home_server_pool", xlat_server_pool, NULL, NULL);
+	xlat_register("home_server_dynamic", xlat_home_server_dynamic, NULL, NULL);
 #endif
 
 	realm_config = rc;
+
+#ifdef HAVE_DIRENT_H
+	if (!rc->dynamic) {
+		if (rc->directory) {
+			WARN("Ignoring 'directory' as dynamic home servers were not configured.");
+		}
+	} else {
+		DIR		*dir;
+		struct dirent	*dp;
+
+		if (!rc->directory) {
+			WARN("Ignoring \"dynamic = true\" due to not set \"directory\" in proxy.conf");
+			return 1;
+		}
+
+		DEBUG2("including files in directory %s", rc->directory);
+
+		dir = opendir(rc->directory);
+		if (!dir) {
+			cf_log_err_cs(config, "Error reading directory %s: %s",
+				      rc->directory, fr_syserror(errno));				      
+			goto error;
+		}
+
+		/*
+		 *	Read the directory, ignoring "." files.
+		 */
+		while ((dp = readdir(dir)) != NULL) {
+			char const *p;
+			char conf_file[PATH_MAX];
+
+			if (dp->d_name[0] == '.') continue;
+
+			/*
+			 *	Check for valid characters
+			 */
+			for (p = dp->d_name; *p != '\0'; p++) {
+				if (isalpha((int)*p) ||
+				    isdigit((int)*p) ||
+				    (*p == '-') ||
+				    (*p == '_') ||
+				    (*p == '.')) continue;
+				break;
+			}
+			if (*p != '\0') continue;
+		
+			snprintf(conf_file, sizeof(conf_file), "%s/%s", rc->directory, dp->d_name);
+			if (home_server_afrom_file(conf_file) < 0) {
+				ERROR("Failed reading home_server from %s - %s",
+				      conf_file, fr_strerror());
+				closedir(dir);
+				goto error;
+			}
+		}
+		closedir(dir);
+	}
+#endif
+
 	return 1;
 }
 
@@ -2512,7 +2745,7 @@ home_server_t *home_server_ldb(char const *realmname,
 		 *	Home servers that are unknown, alive, or zombie
 		 *	are used for proxying.
 		 */
-		if (home->state == HOME_STATE_IS_DEAD) {
+		if (home->state >= HOME_STATE_IS_DEAD) {
 			continue;
 		}
 
@@ -2529,7 +2762,8 @@ home_server_t *home_server_ldb(char const *realmname,
 		 *	came from this server.  Don't re-proxy it
 		 *	there.
 		 */
-		if ((request->listener->type == RAD_LISTEN_DETAIL) &&
+		if (request->listener &&
+		    (request->listener->type == RAD_LISTEN_DETAIL) &&
 		    (request->packet->code == PW_CODE_ACCOUNTING_REQUEST) &&
 		    (fr_ipaddr_cmp(&home->ipaddr, &request->packet->src_ipaddr) == 0)) {
 			continue;
@@ -2637,7 +2871,7 @@ home_server_t *home_server_ldb(char const *realmname,
 		found = pool->fallback;
 
 		WARN("Home server pool %s failing over to fallback %s",
-		      pool->name, found->server);
+		      pool->name, found->virtual_server);
 		if (pool->in_fallback) goto update_and_return;
 
 		pool->in_fallback = true;
@@ -2676,7 +2910,7 @@ home_server_t *home_server_ldb(char const *realmname,
 
 			if (!home) continue;
 
-			if ((home->state == HOME_STATE_IS_DEAD) &&
+			if ((home->state >= HOME_STATE_IS_DEAD) &&
 			    (home->ping_check == HOME_PING_CHECK_NONE)) {
 				home->state = HOME_STATE_ALIVE;
 				home->response_timeouts = 0;
@@ -2736,7 +2970,7 @@ home_server_t *home_server_find(fr_ipaddr_t *ipaddr, uint16_t port,
 #else
 	myhome.proto = IPPROTO_UDP;
 #endif
-	myhome.server = NULL;	/* we're not called for internal proxying */
+	myhome.virtual_server = NULL;	/* we're not called for internal proxying */
 
 	return rbtree_finddata(home_servers_byaddr, &myhome);
 }
@@ -2760,7 +2994,7 @@ home_server_t *home_server_find_bysrc(fr_ipaddr_t *ipaddr, uint16_t port,
 #else
 	myhome.proto = IPPROTO_UDP;
 #endif
-	myhome.server = NULL;	/* we're not called for internal proxying */
+	myhome.virtual_server = NULL;	/* we're not called for internal proxying */
 
 	return rbtree_finddata(home_servers_byaddr, &myhome);
 }
@@ -2785,7 +3019,7 @@ home_server_t *home_server_bynumber(int number)
 
 	memset(&myhome, 0, sizeof(myhome));
 	myhome.number = number;
-	myhome.server = NULL;	/* we're not called for internal proxying */
+	myhome.virtual_server = NULL;	/* we're not called for internal proxying */
 
 	return rbtree_finddata(home_servers_bynumber, &myhome);
 }
@@ -2801,4 +3035,118 @@ home_pool_t *home_pool_byname(char const *name, int type)
 	return rbtree_finddata(home_pools_byname, &mypool);
 }
 
+int home_server_afrom_file(char const *filename)
+{
+	CONF_SECTION *cs, *subcs;
+	char const *p;
+	home_server_t *home;
+
+	if (!realm_config->dynamic) {
+		fr_strerror_printf("Must set \"dynamic = true\" in proxy.conf for dynamic home servers to work");
+		return -1;
+	}
+
+	cs = cf_section_alloc(NULL, "home", filename);
+	if (!cs) {
+		fr_strerror_printf("Failed allocating memory");
+		return -1;
+	}
+
+	if (cf_file_read(cs, filename) < 0) {
+		fr_strerror_printf("Failed reading file %s", filename);
+	error:
+		talloc_free(cs);
+		return -1;
+	}
+
+	p = strrchr(filename, '/');
+	if (p) {
+		p++;
+	} else {
+		p = filename;
+	}
+
+	subcs = cf_section_sub_find_name2(cs, "home_server", p);
+	if (!subcs) {
+		fr_strerror_printf("No 'home_server %s' definition in the file.", p);
+		goto error;
+	}
+
+	home = home_server_afrom_cs(realm_config, realm_config, subcs);
+	if (!home) {
+		fr_strerror_printf("Failed parsing configuration to a home_server structure");
+		goto error;
+	}
+
+	home->dynamic = true;
+
+	if (home->virtual_server || home->dual) {
+		fr_strerror_printf("Dynamic home_server '%s' cannot have 'server' or 'auth+acct'", p);
+		talloc_free(home);
+		goto error;
+	}
+
+#ifdef COA_TUNNEL
+	if (home->recv_coa) {
+		fr_strerror_printf("Dynamic home_server '%s' cannot receive CoA requests'", p);
+		talloc_free(home);
+		goto error;
+	}
+#endif
+
+	if (!realm_home_server_add(home)) {
+		fr_strerror_printf("Failed adding home_server to the internal data structures");
+		talloc_free(home);
+		goto error;
+	}
+
+	return 0;
+}
+
+int home_server_delete(char const *name, char const *type_name)
+{
+	home_server_t *home;
+	int type;
+	char const *p;
+
+	if (!realm_config->dynamic) {
+		fr_strerror_printf("Must set 'dynamic' in proxy.conf for dynamic home servers to work");
+		return -1;
+	}
+
+	type = fr_str2int(home_server_types, type_name, HOME_TYPE_INVALID);
+	if (type == HOME_TYPE_INVALID) {
+		fr_strerror_printf("Unknown home_server type '%s'", type_name);
+		return -1;
+	}
+
+	p = strrchr(name, '/');
+	if (p) {
+		p++;
+	} else {
+		p = name;
+	}
+
+	home = home_server_byname(p, type);
+	if (!home) {
+		fr_strerror_printf("Failed to find home_server %s", p);
+		return -1;
+	}
+
+	if (!home->dynamic) {
+		fr_strerror_printf("Cannot delete static home_server %s", p);
+		return -1;
+	}
+
+	(void) rbtree_deletebydata(home_servers_byname, home);
+	(void) rbtree_deletebydata(home_servers_byaddr, home);
+#ifdef WITH_STATS
+	(void) rbtree_deletebydata(home_servers_bynumber, home);
+#endif
+
+	/*
+	 *	Leak home, and home->cs.  Oh well.
+	 */
+	return 0;
+}
 #endif
